/* client.c */

/* Krzysztof Zmudzinski
   CS 737 Class Project
   Controlled Transfer of Qclic Files
   Date of completion: 5/5/97

   The purpose of this project was to develop an algorithm for
   controlled transfer of Qclic files.  `Controlled' means that the
   user specifies what files should be transferred, how much bandwidth
   should be used to transfer these files, and how much time should
   the transfer take.  The algorithm is responsible for transferring
   the specified files without exceeding the specified bandwidth and time.

   For more info see gqcf.c and the project report at
   http://www.cs.wisc.edu/~zmudzin/737/737.html
*/

#include "gqcf.h"

/* ReceiveData
   
   Client calls this when select indicates that socket has some data;
   data is read and if it is file buffer it is written to the file
   indicated by the packet's header;

   returns -1 for error; 0 otherwise (no errors);
*/

int
ReceiveData(int sockfd,
	    char targetFiles[][MAX_FILE_NAME])
{
  int fd;
  char buff[MAX_BUFF];
  long headerType, long1, long2, long3;

  if (ReadHeader(sockfd, &headerType, &long1, &long2, &long3) < 0) {
    fprintf(stderr, "client error: read header error\n");
    fflush(stderr);
    return -1;
  }
  if (headerType == TRANSFER_QUIT) {
    /* for now just disregard other fields in the packet */
    return 1;
  }
  if (long1 == CREATE) {
    if ((fd = creat(targetFiles[long2], 438)) < 0) {
      fprintf(stderr, "client error: can't create file %s\n",
	      targetFiles[long2]);
      fflush(stderr);
      return -1;
    } else {
      
#ifdef FILEOP
      fprintf(stdout, "client: created file %s %d\n", targetFiles[long2], fd);
      fflush(stdout);
#endif

    }
  } else {
    if ((fd = open(targetFiles[long2], O_WRONLY | O_APPEND)) < 0) {
      fprintf(stderr, 
	      "client error: can't open file %s\n", targetFiles[long2]);
      fflush(stderr);
      return -1;
    } else {
      
#ifdef FILEOP
      fprintf(stdout, "client: opened file %s %d\n", targetFiles[long2], fd);
      fflush(stdout);
#endif
      
    }
  }
    
  if (ReadBytes(sockfd, buff, long3) < 0) {
    fprintf(stderr, "client error: error reading data\n");
    fflush(stderr);
    return -1;
  } else {
    if (write(fd, buff, long3) < 0) {
      fprintf(stderr, "client error: error writing file %d %d\n", fd, errno);
      fflush(stderr);
      return -1;
    }
    
#ifdef FILEIO
    fprintf(stdout, "client: %ld bytes of data received\n", long3);
    fflush(stdout);
#endif
    
  }
  
  close(fd);
  return 0;

}

/* Client

   waits to the socket until somethinbg arrives, it then calls
   ReceiveData; it can also time out after transferTime has expired;
*/

void 
Client(int sockfd, 
       char targetFiles[][MAX_FILE_NAME],
       long transferTime)
{
  struct timeval twait;
  int fd, n;
  fd_set fs;
  int maxfdp1;

  fd = -1; /* assume no file was opened */
  twait.tv_sec = transferTime;

  while (1) {

    /* initialize arguments for select */
    FD_ZERO(&fs);
    FD_SET(sockfd, &fs);

    maxfdp1 = sockfd;

    maxfdp1++;

    if ((n = select(maxfdp1, &fs, 
		    (fd_set *) 0, (fd_set *) 0, &twait) < 0)) {
      fprintf(stderr, "client error: select error\n");
      fflush(stderr);
    }

    if (FD_ISSET(sockfd, &fs)) {
      if (ReceiveData(sockfd, targetFiles) != 0) {
	break;
      }
      continue;
    }

    if (n == 0) {
      fprintf(stderr, "client error: timeout\n");
      fflush(stderr);
      break;
    }

  }

#ifdef CLIENT
  fprintf(stdout, "client: quiting...\n");
  fflush(stdout);
#endif

}

/* SendTransferInfo

   send transfer request to server;

   returns -1 for error; 0 otherwise;
*/

int
SendTransferInfo(int sockfd,
		 char sourceFiles[][MAX_FILE_NAME],
		 long bandwidth,
		 long transferTime,
		 long transferMode)
{
  long fn;
  long reqType;

  if (SendHeader(sockfd, 
		 TRANSFER_REQ, bandwidth, transferTime, transferMode) < 0) {
    fprintf(stderr, "client error: send transfer request error\n");
    fflush(stderr);
    return -1;
  }

  fn = 0;
  while (sourceFiles[fn][0] != '\0') {
    if (sourceFiles[fn + 1][0] == '\0') {
      reqType = LAST_FILE;
    } else {
      reqType = DEFAULT;
    }
      
    if (SendHeader(sockfd, FILE_REQ, 
		     reqType, fn, (long) strlen(sourceFiles[fn])) < 0) {
      fprintf(stderr, "client error: send file request header error\n");
      fflush(stderr);
      return -1;
    }

    if (WriteBytes(sockfd, sourceFiles[fn], strlen(sourceFiles[fn])) < 0) {
      fprintf(stderr, "client error: send file name error\n");
      fflush(stderr);
      return -1;
    }

    fn++;

  }
  return 0;
}

/* GetQclicFiles

   establishes connection with server, sends transfer request
   and starts Client;

   returns -1 for error; 0 otherwise;
*/

int
GetQclicFiles(char *server,
	      char sourceFiles[][MAX_FILE_NAME],
	      char targetFiles[][MAX_FILE_NAME],
	      long bandwidth, 
	      long transferTime, 
	      long transferMode)
{
  
  int sockfd;
  struct sockaddr_in serv_addr;
  struct hostent *host;

  /* used in various places to transfer longs in network byte order */
  longPtr = (long *) malloc(LONGSIZE);

  host = gethostbyname(server);
  if (!host) {
    perror("client: no host info found");
    return -1;
  }

  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  memcpy(&serv_addr.sin_addr.s_addr, host->h_addr, host->h_length);
  serv_addr.sin_port = htons(SERVER_PORT1);

  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0) {
    perror("client: cannot open socket");
    return -1;
  }

  /* try two different ports as the server may be on either of them */
  if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
    serv_addr.sin_port = htons(SERVER_PORT2);
    sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (sockfd < 0) {
      perror("client: cannot open socket");
      return -1;
    }
    if (connect(sockfd, (struct sockaddr *) &serv_addr, 
		sizeof(serv_addr)) < 0) {
      perror("client: cannot connect to server");
      return -1;
    }
  }

  if (SendTransferInfo(sockfd, sourceFiles, 
		       bandwidth, transferTime, transferMode) < 0) {
    fprintf(stderr, "client error: transfer info send error\n");
    fflush(stderr);
    return -1;
  }

  Client(sockfd, targetFiles, transferTime);
  
  shutdown(sockfd, 2);

  return 0;

}

int 
main(int argc, char **argv) 
{
  char sourceFiles[MAX_FILE_NUM][MAX_FILE_NAME];
  char targetFiles[MAX_FILE_NUM][MAX_FILE_NAME];

  if (argc != 5) {
    fprintf(stderr, 
	    "usage: %s server bandwidth transferTime transferMode\n", argv[0]);
    fflush(stderr);
    exit(1);
  }
  if (atoi(argv[4]) < TM_FIRST || atoi(argv[4]) > TM_LAST) {
    fprintf(stderr, "error: wrong value for transferMode\n");
    exit(1);
  }

#ifdef CLIENT
  fprintf(stdout, "client: server = %s\n", argv[1]);
  fprintf(stdout, "client: bandwidth = %ld\n", atol(argv[2]));
  fprintf(stdout, "client: transfer time = %ld\n", atol(argv[3]));
  fprintf(stdout, "client: transfer mode = %ld\n", atol(argv[4]));
  fflush(stdout);
#endif

#ifdef MOBILE
  strcpy(sourceFiles[0], 
	 "mobile000.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile000/mobile000.jpg");*/
  strcpy(sourceFiles[1], 
	 "mobile001.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile001/mobile001.jpg");*/
  strcpy(sourceFiles[2], 
	 "mobile002.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile002/mobile002.jpg");*/
  strcpy(sourceFiles[3], 
	 "mobile003.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile003/mobile003.jpg");*/
  strcpy(sourceFiles[4], 
	 "mobile004.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile004/mobile004.jpg");*/
  strcpy(sourceFiles[5], 
	 "mobile005.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile005/mobile005.jpg");*/
  strcpy(sourceFiles[6], 
	 "mobile006.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile006/mobile006.jpg");*/
  strcpy(sourceFiles[7], 
	 "mobile007.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile007/mobile007.jpg");*/
  strcpy(sourceFiles[8], 
	 "mobile008.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile008/mobile008.jpg");*/
  strcpy(sourceFiles[9], 
	 "mobile009.jpg");
/*	 "/afs/cs.wisc.edu/p/qclic/ims.2/RPI/Mobile/mobile009/mobile009.jpg");*/
#else
  strcpy(sourceFiles[0], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/1.pgm/1.pgm.jpg");
  strcpy(sourceFiles[1], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/2.pgm/2.pgm.jpg");
  strcpy(sourceFiles[2], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/3.pgm/3.pgm.jpg");
  strcpy(sourceFiles[3], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/4.pgm/4.pgm.jpg");
  strcpy(sourceFiles[4], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/5.pgm/5.pgm.jpg");
  strcpy(sourceFiles[5], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/6.pgm/6.pgm.jpg");
  strcpy(sourceFiles[6], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/7.pgm/7.pgm.jpg");
  strcpy(sourceFiles[7], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/8.pgm/8.pgm.jpg");
  strcpy(sourceFiles[8], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/9.pgm/9.pgm.jpg");
  strcpy(sourceFiles[9], 
	 "/afs/cs.wisc.edu/p/qclic/ims.2/orl_faces/s1/10.pgm/10.pgm.jpg");
#endif
  sourceFiles[10][0] = '\0';

  strcpy(targetFiles[0], "0.pgm.jpg");
  strcpy(targetFiles[1], "1.pgm.jpg");
  strcpy(targetFiles[2], "2.pgm.jpg");
  strcpy(targetFiles[3], "3.pgm.jpg");
  strcpy(targetFiles[4], "4.pgm.jpg");
  strcpy(targetFiles[5], "5.pgm.jpg");
  strcpy(targetFiles[6], "6.pgm.jpg");
  strcpy(targetFiles[7], "7.pgm.jpg");
  strcpy(targetFiles[8], "8.pgm.jpg");
  strcpy(targetFiles[9], "9.pgm.jpg");
  targetFiles[10][0] = '\0';

  if (GetQclicFiles(argv[1],
		    sourceFiles, 
		    targetFiles, 
		    atol(argv[2]),
		    atol(argv[3]), 
		    atol(argv[4])) < 0) {
    fprintf(stderr, "client error: transfer failed\n");
    fflush(stderr);
  }
  
  return 0;
}


